local json = require("json")
local const = require("const")
local utils = require("utils2")

utils.check_auth()
utils.check_permission("DeviceManage")

local function list_things()
    local ok, ret, err = utils.dbus_call_edge(
            const.dimu.wkn, const.dimu.obj, const.dimu.inf,
            "getDeviceState")
    if not ok then
        return false, nil, err
    end

    local ok, res = pcall(json.decode, ret[1])
    if not ok or type(res) ~= "table" then
        return false, nil, "not a JSON object retrieved from thing manager"
    end
    if res["code"] ~= 0 then
        return false, nil, tostring(res["message"])
    end
    if not res["params"] or type(res["params"]) ~= "table" or
            not res["params"]["devices"] or type(res["params"]["devices"]) ~= "table" then
        return false, nil, "invalid return value from thing manager"
    end

    local things = {}
    for _, v in ipairs(res["params"]["devices"]) do
        local product_key = v["productKey"]
        local thing = {}
        thing["ProductName"] = ""
        thing["ProductKey"] = product_key or ""
        thing["DeviceName"] = v["deviceName"] or ""
        thing["NickName"] = v["nickName"] or ""
        thing["DriverId"] = v["driverId"] or ""
        thing["DriverName"] = v["driverName"] or ""
        thing["Tags"] = v["tags"] or {}
        thing["Status"] = "Inactivated"
        thing["ConnectTime"] = ""
        thing["Connected"] = false

        if product_key then
            local ok, tsl_str = utils.get_tsl(product_key)
            if ok then
                local tsl
                ok, tsl = pcall(json.decode, tsl_str)
                if ok then
                    if tsl and tsl["profile"] then
                        thing["ProductName"] = tsl["profile"]["productName"] or ""
                    end
                else
                    utils.log_info("decode tsl %s: %s", tsl_str, tsl)
                end
            else
                utils.log_info("get tsl for %s: %s", product_key, tsl_str)
            end
        end

        if v["localOnline"] == "True" then
            thing["Connected"] = true
        end

        local status
        if v["activation"] == "activated" then
            if v["cloudOnline"] == "True" then
                status = "Online"
            else
                status = "Offline"
            end
        elseif v["activation"] == "activationFailed" then
            status = "Failed"
        end
        if status then
            thing["Status"] = status
        end

        if v["device"][3] then
            thing["ConnectTime"] = v["lastLocalOnlineTime"] or 0
        end
        table.insert(things, thing)
    end
    return true, things, nil
end

local request_method = ngx.var.request_method
if request_method == "GET" then
    local ok, things, err = list_things()
    if not ok then
        utils.error_exit(ngx.HTTP_INTERNAL_SERVER_ERROR, err)
    end

    local resp = {}
    resp["Code"] = 200
    resp["Message"] = "success"
    resp["Data"] = {}
    resp["Data"]["Things"] = things
    local str = json.encode(resp)
    ngx.header['Content-Type'] = 'application/json; charset=utf-8'
    ngx.say(str)
else
    return utils.error_exit(ngx.HTTP_NOT_ALLOWED, "request method not allowed: " .. request_method)
end